package com.ichaoj.ams.script.zksync2.swap;


import cn.hutool.core.collection.ListUtil;
import com.ichaoj.ams.script.IScript;
import com.ichaoj.ams.script.SResult;
import com.ichaoj.ams.script.model.AirdropParam;
import com.ichaoj.ams.script.model.AirdropWallet;
import io.zksync.abi.TransactionEncoder;
import io.zksync.crypto.signer.EthSigner;
import io.zksync.crypto.signer.PrivateKeyEthSigner;
import io.zksync.methods.request.Eip712Meta;
import io.zksync.methods.request.Transaction;
import io.zksync.protocol.ZkSync;
import io.zksync.protocol.core.Token;
import io.zksync.protocol.core.ZkBlockParameterName;
import io.zksync.transaction.fee.DefaultTransactionFeeProvider;
import io.zksync.transaction.fee.Fee;
import io.zksync.transaction.fee.ZkTransactionFeeProvider;
import io.zksync.transaction.type.Transaction712;
import lombok.SneakyThrows;
import org.web3j.abi.FunctionEncoder;
import org.web3j.abi.FunctionReturnDecoder;
import org.web3j.abi.TypeReference;
import org.web3j.abi.datatypes.*;
import org.web3j.crypto.Credentials;
import org.web3j.protocol.core.Response;
import org.web3j.protocol.core.methods.response.EthCall;
import org.web3j.protocol.core.methods.response.EthSendTransaction;
import org.web3j.protocol.core.methods.response.TransactionReceipt;
import org.web3j.protocol.http.HttpService;
import org.web3j.utils.Convert;
import org.web3j.utils.Numeric;

import java.math.BigInteger;
import java.util.List;
import java.util.Map;

import static org.web3j.protocol.core.DefaultBlockParameterName.LATEST;

/**
 * @author wrenj
 */
public class ZksyncSwapEthForUsdcScript extends IScript {
    @SneakyThrows
    @Override
    public SResult run(Map<String, AirdropParam> params, AirdropWallet airdropWallet) {
        // Initialize client
        ZkSync zksync = ZkSync.build(new HttpService("https://mainnet.era.zksync.io"));
        Credentials credentials = Credentials.create(airdropWallet.getPrivateKey());
        long chainId = zksync.ethChainId().send().getChainId().longValue();
        EthSigner signer = new PrivateKeyEthSigner(credentials, chainId);
        BigInteger amount = Convert.toWei(params.get("amount").getValue(), Convert.Unit.ETHER).toBigInteger();
        DynamicArray<Address> path = new DynamicArray<>(Address.class, new Address("0x5AEa5775959fBC2557Cc8789bC1bf90A239D9a91"), new Address("0x3355df6D4c9C3035724Fd0e3914dE96A5a83aaf4"));
        DynamicArray<Bool> stable = new DynamicArray<>(Bool.class, new Bool(true), new Bool(false));
        Function getAmountsOut = new Function("getAmountsOut",
                ListUtil.of(
                        new Uint(amount),
                        path,
                        stable),
                ListUtil.of(
                        new TypeReference<DynamicArray<Uint>>() {
                        },
                        new TypeReference<DynamicArray<Bool>>() {
                        },
                        new TypeReference<Uint>() {
                        }));
        EthCall amountOutCall = zksync.ethCall(org.web3j.protocol.core.methods.request.Transaction
                        .createEthCallTransaction(
                                signer.getAddress(),
                                "0x8B791913eB07C32779a16750e3868aA8495F5964",
                                FunctionEncoder.encode(getAmountsOut)),
                ZkBlockParameterName.COMMITTED).send();
        String amountOutCallValue = amountOutCall.getValue();
        List<Type> types = FunctionReturnDecoder.decode(amountOutCallValue, getAmountsOut.getOutputParameters());
        BigInteger amountOut = ((DynamicArray<Uint>) types.get(0)).getValue().get(1).getValue();
        amountOut = amountOut.multiply(BigInteger.valueOf(995)).divide(BigInteger.valueOf(1000));
        Function swapEth = new Function("swapExactETHForTokensSupportingFeeOnTransferTokens", ListUtil.of(
                new Uint(amountOut),
                path,
                new Address(airdropWallet.getAddress()),
                new Uint(BigInteger.valueOf(System.currentTimeMillis() / 1000 + 20 * 60)),
                stable), ListUtil.empty());

        String encodedFunction = FunctionEncoder.encode(swapEth);
        BigInteger nonce = zksync
                .ethGetTransactionCount(signer.getAddress(), ZkBlockParameterName.COMMITTED).send()
                .getTransactionCount();


        String router = "0x8B791913eB07C32779a16750e3868aA8495F5964";

        Transaction estimate = Transaction.createFunctionCallTransaction(
                signer.getAddress(),
                router,
                BigInteger.ZERO,
                BigInteger.ZERO,
                amount,
                encodedFunction
        );
        ZkTransactionFeeProvider feeProvider = new DefaultTransactionFeeProvider(zksync, Token.ETH);

        Transaction712 transaction = new Transaction712(
                chainId,
                nonce,
                feeProvider.getGasLimit(estimate),
                estimate.getTo(),
                estimate.getValueNumber(),
                estimate.getData(),
                BigInteger.valueOf(100000000L),
                feeProvider.getGasPrice(),
                signer.getAddress(),
                estimate.getEip712Meta()
        );

        String signature = signer.getDomain().thenCompose(domain -> signer.signTypedData(domain, transaction)).join();
        byte[] message = TransactionEncoder.encode(transaction, TransactionEncoder.getSignatureData(signature));

        EthSendTransaction sendTransaction = zksync.ethSendRawTransaction(Numeric.toHexString(message)).send();
        Response.Error error = sendTransaction.getError();
        if (error != null) {
            throw new RuntimeException(error.getMessage());
        }
        String sentTransactionHash = sendTransaction.getTransactionHash();

        // You can check the transaction status
        TransactionReceipt receipt = zksync.ethGetTransactionReceipt(sentTransactionHash).send().getTransactionReceipt().get();
        SResult sResult = new SResult();
        sResult.setCurrentBalance(zksync.ethGetBalance(signer.getAddress(), LATEST).send().getBalance().toString());
        sResult.setGas(receipt.getGasUsed().toString());
        sResult.setTxId(receipt.getTransactionHash());


        return sResult;

//        // Initialize client
//        String contractAddress = "0x8B791913eB07C32779a16750e3868aA8495F5964";
//        BigInteger amount = Convert.toWei(params.get("amount").getValue(), Convert.Unit.ETHER).toBigInteger();
//        ZkSync zksync = ZkSync.build(new HttpService("https://mainnet.era.zksync.io"));
//        Credentials credentials = Credentials.create(airdropWallet.getPrivateKey());
//        long chainId = zksync.ethChainId().send().getChainId().longValue();
//        EthSigner signer = new PrivateKeyEthSigner(credentials, chainId);
//        ZkSyncWallet wallet = new ZkSyncWallet(zksync, signer, Token.ETH);
//        DynamicArray<Address> path = new DynamicArray<>(Address.class, new Address("0x5AEa5775959fBC2557Cc8789bC1bf90A239D9a91"), new Address("0x3355df6D4c9C3035724Fd0e3914dE96A5a83aaf4"));
//        DynamicArray<Bool> stable = new DynamicArray<>(Bool.class, new Bool(true), new Bool(false));
//        Function getAmountsOut = new Function("getAmountsOut",
//                ListUtil.of(
//                        new Uint(amount),
//                        path,
//                        stable),
//                ListUtil.of(
//                        new TypeReference<DynamicArray<Uint>>() {
//                        },
//                        new TypeReference<DynamicArray<Bool>>() {
//                        },
//                        new TypeReference<Uint>() {
//                        }));
//        EthCall amountOutCall = zksync.ethCall(org.web3j.protocol.core.methods.request.Transaction
//                        .createEthCallTransaction(
//                                signer.getAddress(),
//                                "0x8B791913eB07C32779a16750e3868aA8495F5964",
//                                FunctionEncoder.encode(getAmountsOut)),
//                ZkBlockParameterName.COMMITTED).send();
//        String amountOutCallValue = amountOutCall.getValue();
//        List<Type> types = FunctionReturnDecoder.decode(amountOutCallValue, getAmountsOut.getOutputParameters());
//        BigInteger amountOut = ((DynamicArray<Uint>) types.get(0)).getValue().get(1).getValue();
//        amountOut = amountOut.multiply(BigInteger.valueOf(995)).divide(BigInteger.valueOf(1000));
//        Function swapEth = new Function("swapExactETHForTokensSupportingFeeOnTransferTokens", ListUtil.of(
//                new Uint(amountOut),
//                path,
//                new Address(airdropWallet.getAddress()),
//                new Uint(BigInteger.valueOf(System.currentTimeMillis() / 1000 + 20 * 60)),
//                stable), ListUtil.empty());
//
//
//        //
////        TransactionReceipt transactionReceipt = wallet.execute(contractAddress, swapEth).send();
////        SResult result = new SResult();
////        result.setTxId(transactionReceipt.getTransactionHash());
////        result.setGas(transactionReceipt.getGasUsed().toString());
////        result.setCurrentBalance(wallet.getBalance().send().toString());
//
//----------------------------------------------------------------------------------

//        BigInteger amount = Convert.toWei(params.get("amount").getValue(), Convert.Unit.ETHER).toBigInteger();
//        DynamicArray<Address> path = new DynamicArray<>(Address.class, new Address("0x5AEa5775959fBC2557Cc8789bC1bf90A239D9a91"), new Address("0x3355df6D4c9C3035724Fd0e3914dE96A5a83aaf4"));
//        DynamicArray<Bool> stable = new DynamicArray<>(Bool.class, new Bool(true), new Bool(false));
//        String router = "0x8B791913eB07C32779a16750e3868aA8495F5964";
//        Web3j web3j = Web3j.build(new HttpService("https://mainnet.era.zksync.io"));
//        Function getAmountsOut = new Function("getAmountsOut",
//                ListUtil.of(
//                        new Uint(amount),
//                        path,
//                        stable),
//                ListUtil.of(
//                        new TypeReference<DynamicArray<Uint>>() {
//                        },
//                        new TypeReference<DynamicArray<Bool>>() {
//                        },
//                        new TypeReference<Uint>() {
//                        }));
//        EthCall amountOutCall = web3j.ethCall(Transaction.createEthCallTransaction(airdropWallet.getAddress(), router, FunctionEncoder.encode(getAmountsOut)), LATEST).send();
//        String amountOutCallValue = amountOutCall.getValue();
//        List<Type> types = FunctionReturnDecoder.decode(amountOutCallValue, getAmountsOut.getOutputParameters());
//        BigInteger amountOut = ((DynamicArray<Uint>) types.get(0)).getValue().get(1).getValue();
//        amountOut = amountOut.multiply(BigInteger.valueOf(995)).divide(BigInteger.valueOf(1000));
//
//        Function swapEth = new Function("swapExactETHForTokensSupportingFeeOnTransferTokens", ListUtil.of(
//                new Uint(amountOut),
//                path,
//                new Address(airdropWallet.getAddress()),
//                new Uint(BigInteger.valueOf(System.currentTimeMillis() / 1000 + 20 * 60)),
//                stable), ListUtil.empty());
//        String swapEthEncode = FunctionEncoder.encode(swapEth);
//        BigInteger gasPrice = web3j.ethGasPrice().send().getGasPrice();
//
//        BigInteger nonce = Web3Util.getNonce(web3j, airdropWallet.getAddress());
//        Transaction transaction = Transaction.createFunctionCallTransaction(
//                airdropWallet.getAddress(),
//                nonce,
//                gasPrice,
//                BigInteger.ZERO,
//                router,
//                amount,
//                swapEthEncode);
//        EthEstimateGas ethEstimateGas = web3j
//                .ethEstimateGas(transaction).send();
//
//        BigInteger gasLimit = ethEstimateGas.getAmountUsed();
//        RawTransaction rawTransaction = RawTransaction.createTransaction(nonce, gasPrice, gasLimit, router, amount, swapEthEncode);
//        return sendTxForEvm(airdropWallet, web3j, rawTransaction);


//        --------------------------------------------------------------------------------------------------------
//        BigInteger amount = Convert.toWei(params.get("amount").getValue(), Convert.Unit.ETHER).toBigInteger();
//        String router = "0x8B791913eB07C32779a16750e3868aA8495F5964";
//////        Web3j web3j = Web3j.build(new HttpService("https://mainnet.era.zksync.io"));
//        ZkSync zkSync = ZkSync.build(new HttpService("https://mainnet.era.zksync.io"));
//        Credentials credentials = Credentials.create(airdropWallet.getPrivateKey());
//        EthSigner signer = new PrivateKeyEthSigner(credentials, zkSync.ethChainId().send().getChainId().longValue());
//        ZkTransactionFeeProvider feeProvider = new DefaultTransactionFeeProvider(zkSync, Token.ETH);
//        ZkSyncTransactionManager transactionManager = new ZkSyncTransactionManager(zkSync, signer, feeProvider);
//
//
//        MuteSwap muteSwap = MuteSwap.load(router, zkSync, transactionManager, new GasProvider());
//        List<String> path = ListUtil.of("0x5AEa5775959fBC2557Cc8789bC1bf90A239D9a91", "0x3355df6D4c9C3035724Fd0e3914dE96A5a83aaf4");
//        List<Boolean> stable = ListUtil.of(true, false);
//        BigInteger amountOut = (BigInteger) muteSwap.getAmountsOut(amount, path, stable).send().get(1);
//        amountOut = amountOut.multiply(BigInteger.valueOf(995)).divide(BigInteger.valueOf(1000));
//        long current = System.currentTimeMillis() / 1000;
//        long deadLine = current + 20 * 60;
//        RemoteFunctionCall<TransactionReceipt> swap = muteSwap.swapExactETHForTokensSupportingFeeOnTransferTokens(
//                amountOut,
//                path,
//                airdropWallet.getAddress(),
//                BigInteger.valueOf(deadLine),
//                stable,
//                amount);
//        TransactionReceipt transactionReceipt = swap.send();
//        ZkSyncWallet wallet;
//
//        String transactionHash = transactionReceipt.getTransactionHash();
//        SResult sResult = new SResult();
//        sResult.setCurrentBalance(null);
//        sResult.setGas(transactionReceipt.getGasUsed().toString());
//        sResult.setTxId(transactionHash);
//
//
//        return sResult;
//        return sendTxForZkEvm(airdropWallet, zkSync, rawTransaction);
    }
}
